/*
 * Copyright (C) 2005-2007 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file compress.c paraslash's dynamic audio range compressor */

/*
 * Used ideas of AudioCompress, (C) 2002-2004  M. Hari Nezumi <magenta@trikuare.cx>
 */

#include "para.h"
#include "compress_filter.cmdline.h"
#include "list.h"
#include "sched.h"
#include "filter.h"
#include "string.h"

/** the size of the output data buffer */
#define COMPRESS_CHUNK_SIZE 40960

/** data specific to the compress filter */
struct private_compress_data {
	/** the current multiplier */
	unsigned current_gain;
	/** points to the configuration data for this instance of the compress filter */
	struct compress_filter_args_info *conf;
	/** minimal admissible gain */
	unsigned min_gain;
	/** maximal admissible gain */
	unsigned max_gain;
	/** number of samples already seen */
	unsigned num_samples;
	/** absolute value of the maximal sample in the current block */
	unsigned peak;
};

static ssize_t compress(char *inbuf, size_t inbuf_len, struct filter_node *fn)
{
	size_t i, length = PARA_MIN((inbuf_len / 2) * 2,
		(fn->bufsize - fn->loaded) / 2 * 2);
	struct private_compress_data *pcd = fn->private_data;
	int16_t *ip = (int16_t *)inbuf, *op = (int16_t *)(fn->buf + fn->loaded);
	unsigned gain_shift = pcd->conf->inertia_arg + pcd->conf->damp_arg,
		mask = (1 << pcd->conf->blocksize_arg) - 1;

	if (!length)
		return 0;
	for (i = 0; i < length / 2; i++) {
		/* be careful in that heat, my dear */
		int sample = *ip++, adjusted_sample;

		if (sample > 0) {
			adjusted_sample = (sample * pcd->current_gain)
				>> gain_shift;
			if (unlikely(adjusted_sample > 32767)) {
				adjusted_sample = 32767;
				pcd->current_gain = (3 * pcd->current_gain +
					(1 << pcd->conf->inertia_arg)) / 4;
				pcd->peak = 0;
			} else
				if (adjusted_sample > pcd->peak)
					pcd->peak = sample;
		} else {
			adjusted_sample = -((-sample * pcd->current_gain)
				>> gain_shift);
			if (unlikely(adjusted_sample < -32768)) {
				adjusted_sample = -32768;
				pcd->current_gain = (3 * pcd->current_gain +
					(1 << pcd->conf->inertia_arg)) / 4;
				pcd->peak = 0;
			} else
				if (-adjusted_sample > pcd->peak)
					pcd->peak = -adjusted_sample;
		}
		*op++ = adjusted_sample;
		if (likely(++pcd->num_samples & mask))
			continue;
		if (pcd->peak < pcd->conf->target_level_arg) {
			if (pcd->current_gain < pcd->max_gain)
				pcd->current_gain++;
		} else {
			if (pcd->current_gain > pcd->min_gain + 1)
				pcd->current_gain -= 2;
		}
//		PARA_DEBUG_LOG("gain: %lu, peak: %d\n", pcd->current_gain,
//			pcd->peak);
		pcd->peak = 0;
//		PARA_INFO_LOG("sample: %lu\n", ABS(sample));
	}
	fn->loaded += length;
	return length;
}

static void close_compress(struct filter_node *fn)
{
	free(fn->private_data);
	free(fn->buf);
}

static void *compress_parse_config(int argc, char **argv)
{
	struct compress_filter_args_info *ret = para_calloc(sizeof(struct compress_filter_args_info));
	if (!compress_cmdline_parser(argc, argv, ret))
		return ret;
	free(ret);
	return NULL;
}

static void open_compress(struct filter_node *fn)
{
	struct private_compress_data *pcd = para_calloc(
		sizeof(struct private_compress_data));
	pcd->conf = fn->conf;
	fn->private_data = pcd;
	fn->bufsize = COMPRESS_CHUNK_SIZE;
	fn->buf = para_malloc(fn->bufsize);
	pcd->current_gain = 1 << pcd->conf->inertia_arg;
	pcd->min_gain = 1 << (pcd->conf->inertia_arg - pcd->conf->aggressiveness_arg);
	pcd->max_gain = 1 << (pcd->conf->inertia_arg + pcd->conf->aggressiveness_arg);
}

/**
 * the init function of the compress filter
 *
 * \param f pointer to the struct to initialize
 */
void compress_init(struct filter *f)
{
	f->open = open_compress;
	f->close = close_compress;
	f->convert = compress;
	f->print_help = compress_cmdline_parser_print_help;
	f->parse_config = compress_parse_config;
}
